// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * AggregationDataStorage.java
 *
 * Created on 9 April 2010
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage.expr;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.swing.JOptionPane;

import fr.unicaen.openmap.dataStorage.OMDataLocation;
import fr.unicaen.openmap.dataStorage.Centroid;
import fr.unicaen.openmap.dataStorage.BoundingBox;
import fr.unicaen.openmap.util.Util;

import fr.unicaen.openmap.jts.JTSFactory;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.simplify.TopologyPreservingSimplifier;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.algorithm.ConvexHull;


import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.omGraphics.OMPoint;
import com.bbn.openmap.omGraphics.OMLine;
import com.bbn.openmap.geo.Geo;
import com.bbn.openmap.proj.Length;
import com.bbn.openmap.util.Debug;

public class SpatialFunction {

    public final static double distanceTolerance = 0.001;

    public static OMDataLocation simplify(OMDataLocation omdl) {
	return simplify(omdl, distanceTolerance);
    }

    public static OMDataLocation simplify(OMDataLocation omdl, double distanceTolerance) {
	return new OMDataLocation(simplify(omdl.getOMGraphic(), distanceTolerance), omdl.getCentroid(), omdl.getBoundingBox());
    }

    public static OMGraphic simplify(OMGraphic omg) {
	return simplify(omg, distanceTolerance);
    }

    public static OMGraphic simplify(OMGraphic omg, double distanceTolerance) {
	Geometry geom =  JTSFactory.OMGraphicToJTSGeometry(omg);
	return JTSFactory.JTSGeometryToOMGraphic(
						 TopologyPreservingSimplifier.simplify(geom, distanceTolerance));
    }

    public static OMDataLocation union(OMDataLocation omdl1, OMDataLocation omdl2) {
	return JTSFactory.JTSGeometryToOMDataLocation(union(omdl1.getOMGraphic(), omdl2.getOMGraphic()));
    }

    public static Geometry union(OMGraphic omg1, OMGraphic omg2) {
	Geometry geom = JTSFactory.OMGraphicToJTSGeometry(omg1);
	return geom.union( JTSFactory.OMGraphicToJTSGeometry(omg2));
    }


    public static OMDataLocation intersection(OMDataLocation omdl1, OMDataLocation omdl2) {
	return JTSFactory.JTSGeometryToOMDataLocation(intersection(omdl1.getOMGraphic(), omdl2.getOMGraphic()));
    }

    public static Geometry intersection(OMGraphic omg1, OMGraphic omg2) {
	Geometry geom = JTSFactory.OMGraphicToJTSGeometry(omg1);
	return geom.intersection( JTSFactory.OMGraphicToJTSGeometry(omg2));
    }


    public static OMDataLocation union(List<OMDataLocation> l) {
	// The construction of one buffer is far more efficient than repeated unions

	OMGraphicList omgl = new OMGraphicList();
	for (OMDataLocation omdl : l) {
	    if (omdl == null)
		continue;
	    OMGraphic omg = omdl.getOMGraphic();
	    if (omg == null)
		continue;
	    omgl.add(omg);
	}
	Geometry res = JTSFactory.OMGraphicToJTSGeometry(omgl).buffer(1E-3);
	return JTSFactory.JTSGeometryToOMDataLocation(res);
    }

    public static OMDataLocation intersection(List<OMDataLocation> l) {
	Geometry inter = null;
	// we must flatten GeomeryCollection
	for (OMDataLocation omdl : l) {
	    if (omdl == null)
		continue;
	    OMGraphic omg = omdl.getOMGraphic();
	    if (omg == null)
		continue;
	    if (inter == null) {
		inter = JTSFactory.OMGraphicToJTSGeometry(omg);
		continue;
	    }
	    inter.intersection(JTSFactory.OMGraphicToJTSGeometry(omg));
	}
	
	return JTSFactory.JTSGeometryToOMDataLocation(inter);
    }

    public static OMDataLocation convexHull(List<OMDataLocation> l) {
	OMGraphicList omgl = new OMGraphicList();
	for (OMDataLocation omdl : l) {
	    if (omdl == null)
		continue;
	    OMGraphic omg = omdl.getOMGraphic();
	    if (omg != null) {
		omgl.add(omg);
	    } else {
		Centroid c = omdl.getCentroid();
		if (c == null)
		    continue;
		// only take the centroid
		omgl.add(new OMPoint(c.getLat(), c.getLon()));
	    }
	}
	ConvexHull convexHull = 
	    new ConvexHull(JTSFactory.extractCoordinates(omgl), JTSFactory.defaultFactory);
	return JTSFactory.JTSGeometryToOMDataLocation(convexHull.getConvexHull());
    }

    public static OMDataLocation centroid(List<OMDataLocation> l) {
	return barycenter(l, null);
    }

    private static enum TypeGeo { ANY, POINT, POLY };
    public static OMDataLocation barycenter(List<OMDataLocation> l, List<Object> weightList) {
	boolean warning = false;
	TypeGeo typeGeo = TypeGeo.ANY;

	// used if typeGeo == TypeGeo.POLY because we need areas of polygons 
	// to compute their centroids
	List<CentroidAreaAndWeight> centroidAreaAndWeights = new ArrayList<CentroidAreaAndWeight>();

	// used if typeGeo == TypeGeo.POINT
	double latSum = 0, lonSum = 0;
	int nbPoint = 0;
	double totalWeight = 0;

	// weight parameter
	Iterator<Object> weightIter = null;
	if (weightList != null)
	    weightIter = weightList.iterator();
	for (OMDataLocation omdl : l) {
	    if (omdl == null)
		continue;
	    OMGraphic omg = omdl.getOMGraphic();
	    double weight = 1;
	    if (weightList != null) {
		if (weightIter.hasNext())
		    weight = Util.parseDouble(weightIter.next());
		else 
		    System.err.println("### Warning: Spatial function barycenter weightList too short");
	    }	
		
	    if (omg instanceof OMPoint) {
		if (typeGeo == TypeGeo.ANY)
		    typeGeo = TypeGeo.POINT;
		if (typeGeo != TypeGeo.POINT) {
		    warning = true;
		    continue;
		}
		nbPoint++;
		totalWeight += weight;
		latSum += weight * ((OMPoint)omg).getLat();
		lonSum += weight * ((OMPoint)omg).getLon();
	    } else if (omg instanceof OMPoly) {
		if (typeGeo == TypeGeo.ANY)
		    typeGeo = TypeGeo.POLY;
		if (typeGeo != TypeGeo.POLY) {
		    warning = true;
		    continue;
		}
		nbPoint++;
		double llPoints[] = ((OMPoly)omg).getRawllpts();
		double a = area(llPoints);
		Centroid c = centroid(llPoints, OMGraphic.RADIANS, a);
		centroidAreaAndWeights.add(new CentroidAreaAndWeight(c, a, weight));
	    } else {
		// only take the centroid
		if (typeGeo == TypeGeo.ANY)
		    typeGeo = TypeGeo.POINT;
		if (typeGeo != TypeGeo.POINT) {
		    warning = true;
		    continue;
		}
		Centroid c = omdl.getCentroid();
		if (c != null) {
		    nbPoint++;
		    totalWeight += weight;
		    latSum += weight * c.getLat();
		    lonSum += weight * c.getLon();
		}
	    }
	} // end for (OMDataLocation omdl : l) 
	if (warning) {
	    System.err.println("*** Centroid Warning: Only the centroid is used for geometry that are not polygons");
//	    JOptionPane.showMessageDialog(null, 
//					  "Only the centroid is used for geometry that are not polygons", 
//					  "Warning", JOptionPane.WARNING_MESSAGE);
	}
	if (nbPoint == 0)
	    return null;
	if (typeGeo == TypeGeo.POINT) {
	    double lon = lonSum/totalWeight;
	    double lat = latSum/totalWeight;
	    Centroid c = new Centroid(lon, lat);
	    Debug.message("DataStorage", "[expr] new centroid : " + c);
	    return new OMDataLocation(new OMPoint(lat, lon), 
				      c,
				      new BoundingBox(lon, lat, lon, lat));
	} else if (typeGeo == TypeGeo.POLY) {
	    if (centroidAreaAndWeights.size() == 0)
		return null;
	    totalWeight = 0;
	    for(CentroidAreaAndWeight caw: centroidAreaAndWeights) {
		double w = caw.weight*caw.area;
		totalWeight += w;
		latSum += w * caw.centroid.getLat();
		lonSum += w * caw.centroid.getLon();
	    }
	    double lon = lonSum / totalWeight;
	    double lat = latSum / totalWeight;
	    Centroid c = new Centroid(lon, lat);
	    Debug.message("DataStorage", "[expr] new centroid : " + c);
	    return new OMDataLocation(new OMPoint(lat, lon), 
				      c,
				      new BoundingBox(lon, lat, lon, lat));
	}	
	return null;
    }

    
    public static double areaInSQKM(OMGraphic omg) {
	return Length.KM.fromRadians(Length.KM.fromRadians(area(omg)));
    }

    public static double area(OMGraphic omg) {
	if (omg instanceof OMPoly)
	    return area(((OMPoly)omg).getRawllpts());
	if (omg instanceof OMGraphicList) {
	    double a = 0;
	    for(OMGraphic om: (OMGraphicList)omg)
		a += area(om);
	    return a;
	}
	System.err.println("*** Area of "+omg+" -> 0");
	return 0;
    }

    public static double area(double llPoints[]) {
	// A = \frac{1}{2} \sum_{i = 0}^{n - 1}( x_i y_{i + 1} - x_{i + 1} y_i)
	double a = 0;
	for(int i = 0; i <= llPoints.length/2-2; i++) 
	    a += llPoints[i*2]*llPoints[(i+1)*2+1] - llPoints[(i+1)*2]*llPoints[i*2+1]; 
	return a/2;
    }
    /** 
     * Compute a centroid of the polygone given by llPoints in lat/lon order.
     * @parameter  lType is either OMGraphic.RADIANS or OMGraphic.DECIMAL_DEGREES
     */
    public static Centroid centroid(double llPoints[], int lType) {
	return centroid(llPoints, lType, null);
    }

    /** 
     * Compute a centroid of the polygone given by llPoints in lat/lon order.
     * @parameter  lType is either OMGraphic.RADIANS or OMGraphic.DECIMAL_DEGREES
     * @parameter _area can be given if known
     */
    public static Centroid centroid(double llPoints[], int lType, Double _area) {
	// x_G=\frac{1}{6 A} \sum_{i = 0}^{n - 1} (x_i + x_{i + 1}) (x_i y_{i + 1} - x_{i + 1} y_i)
	// y_G=\frac{1}{6 A} \sum_{i = 0}^{n - 1} (y_i + y_{i + 1}) (x_i y_{i + 1} - x_{i + 1} y_i)
	double lat = 0;
	double lon = 0;
	for(int i = 0; i <= llPoints.length/2-2; i++) {
	    lat +=  (llPoints[i*2] + llPoints[(i+1)*2]) * (llPoints[i*2]*llPoints[(i+1)*2+1] - llPoints[(i+1)*2]*llPoints[i*2+1]);
	    lon += (llPoints[i*2+1] + llPoints[(i+1)*2+1]) * (llPoints[i*2]*llPoints[(i+1)*2+1] - llPoints[(i+1)*2]*llPoints[i*2+1]);
	} 
	if (_area == null)
	    _area = area(llPoints);
	double a = _area*6;
	lat /= a;
	lon /= a;
	if (lType == OMGraphic.RADIANS) {
	    // Centroid is always in decimal degrees
	    lat = Geo.degrees(lat);
	    lon = Geo.degrees(lon);
	}
	return new Centroid(lon, lat);
    }

    /** angle of the vector in the  lat/lon plane, not the sphere */
    private static double angle(OMLine oml) {
	double coords[] = oml.getLL();
	double lat = coords[2]-coords[0];
	double lon = coords[3]-coords[1];
	// if ( oml.getAttribute(TAG) != FORWARDTAG) {
	//     lat = - lat;
	//     lon = - lon;
	// }
	return Math.atan2(lat, lon);
    }

    /** a1 and a2 are in the interval  [-PI, +PI] 
     *Returns a1 - a2 and normalise the result in order to be in the interval ]0, +2PI]
     */
    private static double subractAngle(double a1, double a2) {
	double angle = a1 - a2; // angle is in [-2 PI, 2PI]
	while ( angle < 1E-10 ) angle += 2*Math.PI;
	// now angle is in ]0, 2PI]
	return angle;
    }

}

class CentroidAreaAndWeight {
    public Centroid centroid;
    public double area;
    public double weight;
    public CentroidAreaAndWeight(Centroid c, double a, double w) {
	centroid = c;
	area = a;
	weight = w;
    }
}
