// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.flowLayer.drawing;

import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.OMFlow;

/** Spline interpolation
 * Adapted from programs by <a href="http://www.cse.unsw.edu.au/~lambert/splines/sources">Tim Lambert</a>
 * @author Jacques.Madelaine@info.unicaen.fr
 * @version Mars 2002, revised July 2003
 **/
public class Spline {

    // Caution: arrays are indexed from 0 to n (i.e. length == n+1)

    /* calculates the natural cubic spline that interpolates
    x[0], x[1], ... x[n]
    The first segment is returned as
    C[0].a + C[0].b*u + C[0].c*u^2 + C[0].d*u^3 0<=u <1
    the other segments are in C[1], C[2], ...  C[n-1] */
    private static Cubic[] calcNaturalCubic(int n, double[] x) {
        double[] gamma = new double[n + 1];
        double[] delta = new double[n + 1];
        double[] D = new double[n + 1];
        int i;
        /* We solve the equation
        [2 1       ] [D[0]]   [3(x[1] - x[0])  ]
        |1 4 1     | |D[1]|   |3(x[2] - x[0])  |
        |  1 4 1   | | .  | = |      .         |
        |    ..... | | .  |   |      .         |
        |     1 4 1| | .  |   |3(x[n] - x[n-2])|
        [       1 2] [D[n]]   [3(x[n] - x[n-1])]

        by using row operations to convert the matrix to upper triangular
        and then back sustitution.  The D[i] are the derivatives at the knots.
         */

        gamma[0] = 1.0f / 2.0f;
        for (i = 1; i < n; i++) {
            gamma[i] = 1 / (4 - gamma[i - 1]);
        }
        gamma[n] = 1 / (2 - gamma[n - 1]);

        delta[0] = 3 * (x[1] - x[0]) * gamma[0];
        for (i = 1; i < n; i++) {
            delta[i] = (3 * (x[i + 1] - x[i - 1]) - delta[i - 1]) * gamma[i];
        }
        delta[n] = (3 * (x[n] - x[n - 1]) - delta[n - 1]) * gamma[n];

        D[n] = delta[n];
        for (i = n - 1; i >= 0; i--) {
            D[i] = delta[i] - gamma[i] * D[i + 1];
        }

        /* now compute the coefficients of the cubics */
        Cubic[] C = new Cubic[n];
        for (i = 0; i < n; i++) {
            C[i] = new Cubic(x[i], D[i], 3 * (x[i + 1] - x[i]) - 2 * D[i] - D[i + 1],
                    2 * (x[i] - x[i + 1]) + D[i] + D[i + 1]);
        }
        return C;
    }
//
//    static public int compute(int n, int[] x, int[] y, int N, int[] X, int[] Y) {
//        if (n < 2) {
//            return 0;
//        }
//        Cubic[] xCubic = calcNaturalCubic(n - 1, x);
//        Cubic[] yCubic = calcNaturalCubic(n - 1, y);
//
//        // very crude technique - just break each segment up into steps lines
//        // Compute the same number of points for each interval
//        // There are n-1 intervals in x
//        // then we compute N/(n-1) interpolate points per interval
//        int k = 0; // index for X and Y vectors
//        int q = (N - 1) / (n - 1);
//        int r = (N - 1) % (n - 1);
//        //	System.out.println(r + " * " + (q+1) + " + " + (n-1-r) + " * " + q);
//
//        // We have: N-1 = (n-1)*q + r
//        // or:      N   = r*(q+1) + (n-1-r)*q + 1
//        // For the r first intervals we compute q+1 points
//        // including the first anchor point.
//        for (int i = 0; i < r; i++) {
//            // System.out.println(k+" 0");
//            X[k] = x[i];
//            Y[k] = y[i];
//            k++;
//            for (int j = 1; j <= q; j++, k++) {
//                double u = j / (double) (q + 1);
//                // System.out.println(k+ " "+ u);
//                X[k] = (int) xCubic[i].eval(u);
//                Y[k] = (int) yCubic[i].eval(u);
//            }
//        }
//
//        // For the n-1-r last intervals we compute q points => q intervals
//        for (int i = r; i < n - 1; i++) {
//            // System.out.println(k+" 0");
//            X[k] = x[i];
//            Y[k] = y[i];
//            k++;
//            for (int j = 1; j < q; j++, k++) {
//                double u = j / (double) q;
//                // System.out.println(k+ " " + u);
//                X[k] = (int) xCubic[i].eval(u);
//                Y[k] = (int) yCubic[i].eval(u);
//            }
//        }
//
//        X[k] = x[n - 1];
//        Y[k] = y[n - 1];
//        k++;
//
//        return k;
//
//    }
    static public int compute(int n, double[] x, double[] y, int N, double[] X, double[] Y) {
        if (n < 2) {
            return 0;
        }

        Cubic[] xCubic = calcNaturalCubic(n - 1, x);
        Cubic[] yCubic = calcNaturalCubic(n - 1, y);

        // very crude technique - just break each segment up into steps lines
        // Compute the same number of points for each interval
        // There are n-1 intervals in x
        // then we compute N/(n-1) interpolate points per interval
        int k = 0; // index for X and Y vectors
        int q = (N - 1) / (n - 1);
        int r = (N - 1) % (n - 1);
        //	System.out.println(r + " * " + (q+1) + " + " + (n-1-r) + " * " + q);

        // We have: N-1 = (n-1)*q + r
        // or:      N   = r*(q+1) + (n-1-r)*q + 1
        // For the r first intervals we compute q+1 points
        // including the first anchor point.
        for (int i = 0; i < r; i++) {
            // System.out.println(k+" 0");
            X[k] = x[i];
            Y[k] = y[i];
            k++;
            for (int j = 1; j <= q; j++, k++) {
                double u = j / (double) (q + 1);
                // System.out.println(k+ " "+ u);
                X[k] = xCubic[i].eval(u);
                Y[k] = yCubic[i].eval(u);
            }
        }

        // For the n-1-r last intervals we compute q points => q intervals
        for (int i = r; i < n - 1; i++) {
            // System.out.println(k+" 0");
            X[k] = x[i];
            Y[k] = y[i];
            k++;
            for (int j = 1; j < q; j++, k++) {
                double u = j / (double) q;
                // System.out.println(k+ " " + u);
                X[k] = xCubic[i].eval(u);
                Y[k] = yCubic[i].eval(u);
            }
        }

        X[k] = x[n - 1];
        Y[k] = y[n - 1];
        k++;

        return k;
    }
//    public static double[] computeRadialFromSpline(int[][] spline, int x, int y, int width) {
//        int[] xs = spline[0];
//        int[] ys = spline[1];
//        int index = 0;
//        for (int i = 0; i < xs.length; i++) {
//            if (xs[i] == x && ys[i] == y) {
//                index = i;
//                break;
//            }
//        }
//
//        double[] radialVector;
//
//        if (index == xs.length - 1) {
//            radialVector = OMFlow.computeRadial(xs[index - 1], ys[index - 1],
//                    xs[index], ys[index]);
//        } else {
//            radialVector = OMFlow.computeRadial(xs[index], ys[index],
//                    xs[index+ 1], ys[index + 1]);
//        }
//        return radialVector;
//
//    }
    /** just for test */
//    static public void main(String argv[]) {
//        double x[] = {0, 1, 2, 2};
//        double y[] = {0, 1, 0, 1};
//        int n = x.length;
//
//        int N = 40;
//        double xx[], yy[];
//        xx = new double[N];
//        yy = new double[N];
//        int nb = Spline.compute(n, x, y, N, xx, yy);
//        System.out.println(n + " => " + N + "(" + nb + ")");
//        for (int i = 0; i < nb; i++) {
//            System.out.println(xx[i] + " \t" + yy[i]);
//        }
//    }
}

/** this class represents a cubic polynomial */
class Cubic {

    double a, b, c, d;         /* a + b*u + c*u^2 +d*u^3 */


    public Cubic(double a, double b, double c, double d) {
        this.a = a;
        this.b = b;
        this.c = c;
        this.d = d;
    }

    /** evaluate cubic */
    public double eval(double u) {
        return (((d * u) + c) * u + b) * u + a;
    }
}
////** Spline interpolation
/// * Adapted from <a href="http://www.nr.com">Numerical Recipes </a>
/// * @author Jacques.Madelaine@info.unicaen.fr
/// * @version Mars 2002, revised July 2003
/// **/
///
///package fr.unicaen.openmap.layer.flowLayer.drawing;
///
///import java.util.*;
///
///
///public class Spline {
///    private int nba;		// the number of anchor points
///    private double xa[], ya[]; 	// the anchor points
///    private double y2a[]; 	// the 2d derivative values
///
///    /**
///     * Ctor with n points (x[i], y[i]) i=0,nba-1
///     * the arrays x and y should have the same length
///     **/
///    public Spline(double x[], double y[]){
///	this(x, y, 1e30F, 1e30F);
///    }
///
///    /**
///     * Ctor with n points (x[i], y[i])
///     * and the value of the first derivative at (x[0],y[0]) and (x[n-1],y[n-1]) 
///     **/
///    public Spline(double x[], double y[], double yp0, double ypn_1) {
///	int n = x.length;
///	nba = n;
///	xa = x;
///	ya = y; 
///	y2a = new double[n+1];
///	double u[] = new double[n];
///
///	// compute y2a
///	if (yp0 > 0.99e30)
///	    y2a[0] = u[0] = (double)0.0;
///	else {
///	    y2a[0] = (double)-0.5;
///	    u[0] = ((double)3.0/(xa[1]-xa[0])) * ((ya[1]-ya[0]) / (xa[1]-xa[0]) - yp0);
///	}
///	for(int i = 1; i < n-1; i++) {
///	    double sig = (xa[i]-xa[i-1]) / (xa[i+1]-xa[i-1]);
///	    double p = sig * y2a[i-1] + (double)2.0;
///	    y2a[i] = (sig - 1.0F) / p;
///	    u[i] = (ya[i+1]-ya[i])/(xa[i+1]-xa[i]) - (ya[i]-ya[i-1])/(xa[i]-xa[i-1]);
///	    u[i] = (6.0F*u[i]/(xa[i+1]-xa[i-1]) - sig*u[i-1]) / p;
///	}
///	
///	double qn_1;
///	
///	if (ypn_1 > 0.99e30)
///	    qn_1 = u[n-1] = 0;
///	else {
///	    qn_1 = 0.5F;
///	    u[n-1] = (3.0F / (xa[n-1]-xa[n-2])) * (ypn_1-(ya[n-1]-ya[n-2])/(xa[n-1]-xa[n-2]));
///	}
///	y2a[n-1] = (u[n-1]-qn_1*u[n-2])/(qn_1*y2a[n-2]+1.0F);
///	for(int k = n-1; k >= 0; k--)
///	    y2a[k] = y2a[k]*y2a[k+1]+u[k];
///    }
///      
///    /**
///     * Computes the interpolated value y for the given value of x 
///     * @return the interpolated value
///     */
///    public double splint(double x){
///	int klo = 0;
///	int khi = nba-1;
///	while ((khi - klo) > 1) {
///	    int k = (khi+klo) >> 1;
///	    if (xa[k] > x)
///		khi = k;
///	    else
///		klo = k;
///	}
///	return splint(x, khi, klo);
///    }
///
///    /**
///     * Computes the interpolated value y for the given value of x 
///     * that verifies xa[klo] <= x < xa[khi]  
///     * @returns the interpolated value
///     */
///    private double splint(double x, int khi, int klo) {
///	double h = xa[khi] - xa[klo];
///	if (h == 0.0) {
///	    //  cerr << "bad xa input" << endl;
///	    return ya[khi];
///	}
///	double a = (xa[khi]-x) / h;
///	double b = (x-xa[klo]) / h;
///	return a*ya[klo] + b*ya[khi] + ((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi]) * (h*h)/6.0F;
///    }
///
///    /**
///     * Computes n interpolated points (x[i], y[i])
///     * with the same number of points for each interval 
///     * of anchors points xa[i] xa[i+1]
///     * @return the number of computed points
///     */
///    public int splint(double x[], double y[]) {
///	return splint(x, y, true);
///    }
///
///    /**
///     * Computes n interpolated points (x[i], y[i])
///     * uniformely spreaded between xa[0] and xa[n-1]
///     * or with the same number of points for each interval of anchors points xa[i] xa[i+1]
///     * if equalNuberOfPointsPerInterval is true
///     * @return the numbre of computed points
///     */
///    public int splint(double x[], double y[], boolean equalNuberOfPointsPerInterval) {
///	int n = x.length;
///	if (equalNuberOfPointsPerInterval) { 
///	    //// Compute the same number of points for each interval
///	    // There are nba-1 intervals in xa 
///	    // then we compute n/(nba-1) interpolate points per interval 
///	    int k = 0; // index for x and y vectors
///	    int q = (n-1)/(nba-1);
///	    int r = (n-1)%(nba-1);
///	    // We have: n = r*(q+1)+(nba-r)*q + 1
///  
///	    // For the r first intervals we compute q+1 points => q+1 intervals
///	    // with the point of the next interval.
///	    for(int i = 0; i < r; i++) {
///		x[k] = xa[i]; y[k] = ya[i]; k++;
///		double h = (xa[i+1]-xa[i])/(q+1);
///		double hh = h;
///		for (int j = 1; j < q+1; j++, hh+=h, k++)
///		    y[k] = splint(x[k]=xa[i]+hh, i, i+1);
///	    }
///	    // For the nba-r last intervals we compute q points => q intervals
///	    for(int i = r; i < nba-1; i++) {
///		x[k] = xa[i]; y[k] = ya[i]; k++;
///		double h = (xa[i+1]-xa[i])/q;
///		double hh = h;
///		for (int j = 1; j < q; j++, hh+=h, k++)
///		    y[k] = splint(x[k]=xa[i]+hh,i, i+1);   
///	    }
///	    
///	    x[k] = xa[nba-1]; y[k] = ya[nba-1]; k++;
///	    return k;
///	}
///	//// the interpolated points are uniformely spreaded between xa[0] and xa[n-1]
///	double h = (xa[nba-1]-xa[0])/(n-1);
///	double hh = 0;
///	int i = 0; // x[k] is in the interval (xa[i], xa[i+1])
///	for(int k = 0; k < n-1; k++, hh+=h) {
///	    x[k]=xa[0]+hh;
///	    while (xa[i+1] < x[k])
///		i++;
///	    if (! (xa[i] <= x[k] && x[k] <= xa[i+1]))
///		System.err.println("** strange" + k + " " + i + " " + x[k] + "  " + xa[i]);
///	    y[k] = splint(x[k],i, i+1);
///	}
///	x[n-1] = xa[nba-1]; y[n-1] = ya[nba-1];
///	return n;
///    }
///
///    static public void main(String argv[]) {
///	double x[] = { 0, 1, 2 };
///	double y[] = { 0, 1, 0 };
///	int n = x.length;
///	Spline S = new Spline(x, y);
///        int N = 200;
///	double xx[], yy[];
///	xx = new double[N];
///	yy = new double[N];
///	S.splint(xx, yy);
///	for (int i = 0; i < N; i++)
///	    System.out.println(xx[i] + " \t" + yy[i]);
///    }
///}

