// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu;

import fr.unicaen.openmap.event.RefreshableGUI;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Paint;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.util.*;
import javax.swing.*;

import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.util.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.legend.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;
import fr.unicaen.openmap.util.Util;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.*;
import java.awt.Dimension;

/**
 * Proportional Symbol (Polygon, Disc, Pictogram) visualization mode.
 */
public class ProportionalSymbol extends OffsetableRaw implements ResizerListener {

    public static final String showNullAsPointPropertyString = "showNullAsPoint";
    public static final String modePropertyString = "mode";
    public static final String nbEdgesPropertyString = "nbEdges";
    public static final String starCoefPropertyString = "starCoef";
    public static final String startAnglePropertyString = "startAngle";
    public static final String pictogramPropertyString = "pictogram";

    public enum type {

        DISC, POLY, PICTO
    };
    private type mode;
    private String pictogram = null;
    protected SymbolResizer resizer =
            new SymbolResizer(new DataSquareRootColumnValueComputer(""));

    public String getDataColumnName() {
        return resizer.getDataColumnName();
    }

    public void setDataColumnName(String name) {
        resizer.setDataColumnName(name);
        refresh();
    }
    //protected Offseter offset = new Offseter();
    public static int nbDiscEdges = 21;
    public static int nbPolyEdges = 3;
    protected int nbEdges = 21;
    protected int starCoef = 0;
    protected int startAngle = 45; // in degrees
    private boolean showNullAsPoint;

    protected boolean getShowNullAsPoint() {
        return showNullAsPoint;
    }

    protected void setShowNullAsPoint(boolean b) {
        if (showNullAsPoint != b) {
            showNullAsPoint = b;
            refresh();
        }
    }
    protected MinMaxValues minMaxValues;

    public ProportionalSymbol(ThematicDataDisplayLayer layer, String name) {
        this(layer, name, i18n.get(ProportionalSymbol.class,
                "proportionalSymbol", " Proportional symbol"));
    }

    public ProportionalSymbol(ThematicDataDisplayLayer layer, String name,
            String typeName) {
        super(layer, name, typeName);
        resizer.addResizerListener(this);
        //colors.setFillMode(true);
        mode = type.DISC;
        //offseter.addOffseterListener(this);
        interactive = true;
    }

    @Override
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        setShowNullAsPoint(Util.isTrue(properties.getProperty(prefix + showNullAsPointPropertyString, "false")));
        resizer.setProperties(prefix, properties);
        setMode(Enum.valueOf(type.class, properties.getProperty(prefix + modePropertyString, "DISC")));
        setNbEdges(PropUtils.intFromProperties(properties, prefix + nbEdgesPropertyString, getNbEdges()));
        setStartAngle(PropUtils.intFromProperties(properties, prefix + startAnglePropertyString, getStartAngle()));
        setStarCoef(PropUtils.intFromProperties(properties, prefix + starCoefPropertyString, getStarCoef()));
        pictogram = properties.getProperty(prefix + pictogramPropertyString, "");
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        properties.put(prefix + showNullAsPointPropertyString, new Boolean(getShowNullAsPoint()).toString());
        properties = resizer.getProperties(prefix, properties);
        properties.put(prefix + modePropertyString, mode.name());
        properties.put(prefix + nbEdgesPropertyString, Integer.toString(getNbEdges()));
        properties.put(prefix + startAnglePropertyString, Integer.toString(getStartAngle()));
        properties.put(prefix + starCoefPropertyString, Integer.toString(getStarCoef()));
        if (pictogram != null) {
            properties.put(prefix + pictogramPropertyString, pictogram);
        }

        return properties;
    }

    private static Color negative(Color c) {
        return new Color(255 - c.getRed(), 255 - c.getGreen(), 255 - c.getBlue(), c.getAlpha());
    }

    @Override
    public OMGraphicList generateGraphics(DataStorageIterator dsi) {
        try {
            resizer.setMinMaxValue(dsi);
            OMSizeableGraphicList omglPoly = new OMSizeableGraphicList<OMRegularPolygon>();
            mouseControlPointsListener.clearControlPoints();
            float startAngleInRadians = getStartAngleInRadians();
            boolean offVis = offsetsOn ? offseter.isVisible() : false;
            boolean offLineVis = offsetsOn ? offseter.isLineVisible() : false;

            for (int i = 0; dsi.hasNext(); i++) {
                DataObject data = dsi.next();
                OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
                if (omdl == null) {
                    continue;
                }

                Centroid centroid = omdl.getCentroid();
                if (centroid == null) {
                    continue;
                }

                double radius = resizer.getSize(data);
                float lat = (float) centroid.getLat();
                float lon = (float) centroid.getLon();
                int offX = 0;
                int offY = 0;
                String key = "";
                if (offsetsOn) {
                    key = offseter.buildKey(data);
                    offX = offseter.getX(key);
                    offY = offseter.getY(key);
                }
                OMGraphic p;
                if (getShowNullAsPoint() && ((int) (2 * Math.abs(radius))) < 2) {
                    p = new OMPoint((float) centroid.getLat(), (float) centroid.getLon(), 1);
                    omglPoly.add(p);
                    //System.out.println("RADIUS = " + radius);

                } else {
                    if (mode != type.PICTO) {
                        if (nbEdges < 21 || starCoef > 0) { //draw poly
                            p = new OMRegularPolygon(this, key, lat,
                                    lon, nbEdges, (float) Math.abs(radius),
                                    offX, offY, startAngleInRadians,
                                    starCoef, resizer.getString(data), offVis, offLineVis);
                            mouseControlPointsListener.addOMControlPoints(((OMRegularPolygon)p).controlPoints);
                        } else {
                            //draw circle. Beware use the diameter!
                            //System.out.println("NEW OMDISC DE DIAMETRE:" + (int) (Math.abs(radius) * 2));
                            p = new OMDisc(this, key, (float) lat,
                                    lon, offX, offY, (int) (Math.abs(radius) * 2), resizer.getString(data),
                                    offVis, offLineVis);
                           mouseControlPointsListener.addOMControlPoints(((OMDisc)p).controlPoints);
                        }
                    } else {                                // draw picto
                        if (radius <= 2) { // pb with neg radius
                            radius = 2;
                        }
                        p = new OMPictogram(this, key, lat, lon,
                                offX,
                                offY,
                                OMPictogram.scaleImageIcon(Util.loadIcon(pictogram), (int) Math.abs(radius) * 2), offVis, offLineVis);
                                mouseControlPointsListener.addOMControlPoints(((OMPictogram)p).controlPoints);

                    }
                    omglPoly.add(p);
//                    if (p instanceof OMDisc) {
//                        OMDisc d = (OMDisc) p;
//                        /*if(offsetsOn)
//                            d.addOffsetableListener(offseter);*/
//                        mouseControlPointsListener.addOMControlPoints(d.controlPoints);
//                    }
                }               

                omglPoly.add(p);
                p.setLinePaint(drawingAttributes.getLinePaint());
                p.setFillPaint(radius >= 0 ? drawingAttributes.getFillPaint() : negative((Color) drawingAttributes.getFillPaint()));
                p.setStroke(drawingAttributes.getStroke());
            }
            omglPoly.sort();

            initMinMaxValues(omglPoly); // used for the legend

            return omglPoly;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return new OMGraphicList();
    }

    public boolean isReady() {
        if (resizer.getIsFixedSize()) {
            return super.isReady();
        } else {
            return super.isReady() &&
                    !"".equals(resizer.getDataColumnName()) &&
                    resizer.getDataColumnName() != null;
        }
    }

    public String getMissingInfos() {
        String mess = getName() + ":\n";
        if (geometryColumnName.equals("")) {
            mess += i18n.get(this, "geometryColumnName", "Geometry column name") + "\n";
        }
        if (resizer.getDataColumnName().equals("")) {
            mess += i18n.get(ResizerPanel.class, "dataColumnNameLabel", "Data Column Name") + "\n";
        }
        return mess;

    }

    protected JPanel dataPanel; // protected attribute in order for derived class to keep this one!

    public JComponent getGUI(final DataStorageIterator dsi) {
        thematicGUI = (ThematicGUI) super.getGUI(dsi);
        JCheckBox showNullAsPointCB = new JCheckBox(i18n.get(this, "showNullAsPoint", "Show Null As Point"));
        showNullAsPointCB.setSelected(getShowNullAsPoint());
        showNullAsPointCB.addItemListener(new ItemListener() {

            public void itemStateChanged(ItemEvent e) {
                setShowNullAsPoint(((AbstractButton) (e.getItemSelectable())).isSelected());
            }
        });


        ThematicConfPanel resizerConfPanel = (ThematicConfPanel) resizer.getGUI(dsi);

        JPanel wrapResize = new JPanel();
        wrapResize.add(resizerConfPanel);
        thematicGUI.add(wrapResize);
        thematicGUI.addThematicConfigurationPanel(resizerConfPanel);
        JPanel pShowNull = new JPanel(new BorderLayout());
        pShowNull.add(showNullAsPointCB, BorderLayout.WEST);
        pShowNull.setBorder(BorderFactory.createEmptyBorder(0, 50, 0, 0));
        resizer.getProportionalGUI().add(pShowNull);


        ThematicConfPanel proportionalConf = (ThematicConfPanel) getProportionalSymbolGUI();

        JPanel wrapProp = new JPanel();
        wrapProp.add(proportionalConf);
        thematicGUI.add(wrapProp);
        thematicGUI.addThematicConfigurationPanel(proportionalConf);


        return thematicGUI;
    }
    protected ProportionalSymbolPanel proportionalSymbolPanel = null;

    public ThematicConfPanel getProportionalSymbolGUI() {
        if (proportionalSymbolPanel == null) {
            proportionalSymbolPanel = new ProportionalSymbolPanel(this);
        }
        ThematicConfPanel cp = ThematicConfPanel.createThematicConfPanel(
                i18n.get(ProportionalSymbolPanel.class, "title", "drawing"));
        cp.setContent(proportionalSymbolPanel);
        return cp;
    }

    public void setStartAngle(int a) {
        startAngle = a;
        refresh();
    }

    public void setStarCoef(int c) {
        starCoef = c;
        refresh();
    }

    public void setNbEdges(int n) {
        nbEdges = n;
        refresh();
    }

    /** in order for MultiRegularPolygons to overload this method */
    public void setElemRadius(int r) {
        refresh();
    }

    public int getStartAngle() {
        return startAngle;
    }

    public float getStartAngleInRadians() {
        return (float) (startAngle * Math.PI / 180.0);
    }

    public int getStarCoef() {
        return starCoef;
    }

    public int getNbEdges() {
        return nbEdges;
    }

    /** in order for MultiRegularPolygons to overload this method */
    public int getElemRadius() {
        return 20;
    }

    protected void initMinMaxValues(OMGraphicList omgl) {
	if (omgl == null || omgl.size() == 0) {
	    minMaxValues = new MinMaxValues(0, 0);
	    return;
	}

	double min = 0, max = 0;
	OMGraphic omg = omgl.getOMGraphicAt(0);
	try {
	    min = Util.parseDouble(((HaveWidthAndData) omgl.getOMGraphicAt(0)).getData());
	} catch (java.lang.ClassCastException e) {
	}
	try {
	    max = Util.parseDouble(((HaveWidthAndData) omgl.getOMGraphicAt(omgl.size() - 1)).getData());
	} catch (java.lang.ClassCastException e) {
	}
	minMaxValues = new MinMaxValues(min, max);
    }

    @Override
    public OMResizeableGraphic getLegend(Refreshable r, MouseControlPointsListener ml) {
        if (minMaxValues == null) {
            return null;
        }
        OMResizeableFrame frame = (mode != type.PICTO && nbEdges > 20)
                ? new OMResizeableDiscLegend(r, ml, 0, 0, resizer, minMaxValues, getDataColumnName(),
                drawingAttributes.getLinePaint(), drawingAttributes.getFillPaint(), getFullName())
                : new OMResizeableRegularPolygonLegend(r, ml, 0, 0, resizer, minMaxValues,
                nbEdges, getStartAngleInRadians(),
                starCoef, getDataColumnName(),
                drawingAttributes.getLinePaint(), drawingAttributes.getFillPaint(), getFullName());
        frame.setVisuFrame(true);
        return frame;
    }

    public void refreshGUI(DataStorageIterator dsi) {
        super.refreshGUI(dsi);
        ((RefreshableGUI) resizer).refreshGUI(dsi);
    }

    public void resizerChanged(Resizer resizer) {

        if (this.resizer == resizer) {
            refresh();
        }
    }

    public void setMode(type mode) {
        this.mode = mode;
    }

    public type getMode() {
        return this.mode;
    }

    public void setPictogram(String pictogram) {
        this.pictogram = pictogram;
    }

    public SymbolResizer getResizer() {
        return resizer;
    }
}
